﻿using System;

namespace TestProject
{
    /// <summary>
    /// 복소수
    /// </summary>
    public class Complex
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 실수 - RealNumber

        /// <summary>
        /// 실수
        /// </summary>
        public double RealNumber { get; set; }

        #endregion
        #region 허수 - ImaginaryNumber

        /// <summary>
        /// 허수
        /// </summary>
        public double ImaginaryNumber { get; set; }

        #endregion

        #region 크기 - Magnitude

        /// <summary>
        /// 크기
        /// </summary>
        public double Magnitude
        {
            get
            {
                return Math.Sqrt(RealNumber * RealNumber + ImaginaryNumber * ImaginaryNumber);
            }
        }

        #endregion
        #region 제곱근 - SquareRoot

        /// <summary>
        /// 제곱근
        /// </summary>
        public Complex SquareRoot
        {
            get
            {
                double newRealNumber      = Math.Sqrt((Magnitude + RealNumber) / 2.0);
                double newImaginaryNumber = Math.Sign(ImaginaryNumber) * Math.Sqrt((Magnitude - RealNumber) / 2.0);

                return new Complex(newRealNumber, newImaginaryNumber);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - Complex(realNumber, imaginaryNumber)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="realNumber">실수</param>
        /// <param name="imaginaryNumber">허수</param>
        public Complex(double realNumber, double imaginaryNumber)
        {
            RealNumber      = realNumber;
            ImaginaryNumber = imaginaryNumber;
        }

        #endregion
        #region 생성자 - Complex(realNumber)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="realNumber">실수</param>
        public Complex(double realNumber) : this(realNumber, 0)
        {
        }

        #endregion
        #region 생성자 - Complex()

        /// <summary>
        /// 생성자
        /// </summary>
        public Complex() : this(0, 0)
        {
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 연산자 + 재정의하기 - +(complex1, complex2)

        /// <summary>
        /// 연산자 + 재정의하기
        /// </summary>
        /// <param name="complex1">복소수 1</param>
        /// <param name="complex2">복소수 2</param>
        /// <returns>복소수</returns>
        public static Complex operator +(Complex complex1, Complex complex2)
        {
            return new Complex(complex1.RealNumber + complex2.RealNumber, complex1.ImaginaryNumber + complex2.ImaginaryNumber);
        }

        #endregion
        #region 연산자 - 재정의하기 - -(complex)

        /// <summary>
        /// 연산자 - 재정의하기
        /// </summary>
        /// <param name="complex">복소수</param>
        /// <returns>복소수</returns>
        public static Complex operator -(Complex complex)
        {
            return new Complex(-complex.RealNumber, -complex.ImaginaryNumber);
        }

        #endregion
        #region 연산자 - 재정의하기 - -(complex1, complex2)

        /// <summary>
        /// 연산자 - 재정의하기
        /// </summary>
        /// <param name="complex1">복소수 1</param>
        /// <param name="complex2">복소수 2</param>
        /// <returns>복소수</returns>
        public static Complex operator -(Complex complex1, Complex complex2)
        {
            return complex1 + (-complex2);
        }

        #endregion
        #region 연산자 * 재정의하기 - *(complex1, complex2)

        /// <summary>
        /// 연산자 * 재정의하기
        /// </summary>
        /// <param name="complex1">복소수 1</param>
        /// <param name="complex2">복소수 2</param>
        /// <returns>복소수</returns>
        public static Complex operator *(Complex complex1, Complex complex2)
        {
            return new Complex
            (
                complex1.RealNumber * complex2.RealNumber      - complex1.ImaginaryNumber * complex2.ImaginaryNumber,
                complex1.RealNumber * complex2.ImaginaryNumber + complex1.ImaginaryNumber * complex2.RealNumber
            );
        }

        #endregion
        #region 연산자 / 재정의하기 - /(complex1, complex2)

        /// <summary>
        /// 연산자 / 재정의하기
        /// </summary>
        /// <param name="complex1">복소수 1</param>
        /// <param name="complex2">복소수 2</param>
        /// <returns>복소수</returns>
        public static Complex operator /(Complex complex1, Complex complex2)
        {
            return complex1 * complex2.Inverse();
        }

        #endregion

        #region 암시적 변환 정의하기 - Complex(value)

        /// <summary>
        /// 암시적 변환 정의하기
        /// </summary>
        /// <param name="value">값</param>
        public static implicit operator Complex(double value)
        {
            return new Complex(value);
        }

        #endregion

        #region 파싱하기 - Parse(text)

        /// <summary>
        /// 파싱하기
        /// </summary>
        /// <param name="text">텍스트</param>
        /// <returns>복소수</returns>
        public static Complex Parse(string text)
        {
            text = text.Replace(" ", "");
            text = text.ToLower().Replace("i", "");

            char[] signCharacterArray = { '+', '-' };

            int position = text.IndexOfAny(signCharacterArray);

            if(position == 0)
            {
                position = text.IndexOfAny(signCharacterArray, 1);
            }

            double realNumber      = double.Parse(text.Substring(0, position));
            double imaginaryNumber = double.Parse(text.Substring(position));

            return new Complex(realNumber, imaginaryNumber);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 극좌표 값 구하기 - GetPolarValue(r, theta)

        /// <summary>
        /// 극좌표 값 구하기
        /// </summary>
        /// <param name="r">R</param>
        /// <param name="theta">세타</param>
        public void GetPolarValue(out double r, out double theta)
        {
            r = Math.Sqrt(RealNumber * RealNumber + ImaginaryNumber * ImaginaryNumber);

            theta = Math.Atan2(ImaginaryNumber, RealNumber);
        }

        #endregion
        #region 거듭제곱 구하기 - GetPower(power)

        /// <summary>
        /// 거듭제곱 구하기
        /// </summary>
        /// <param name="power">거듭제곱 수</param>
        /// <returns>거듭제곱 값</returns>
        /// <remarks>De Moivre 방정식을 사용하여 숫자를 거듭제곱한다.</remarks>
        public Complex GetPower(double power)
        {
            GetPolarValue(out double r, out double theta);

            double rToPower           = Math.Pow(r, power);
            double newRealNumber      = rToPower * Math.Cos(power * theta);
            double newImaginaryNumber = rToPower * Math.Sin(power * theta);

            return new Complex(newRealNumber, newImaginaryNumber);
        }

        #endregion
        #region 반전하기 - Inverse()

        /// <summary>
        /// 반전하기
        /// </summary>
        /// <returns></returns>
        public Complex Inverse()
        {
            double magnitude = Magnitude;

            return new Complex(RealNumber / magnitude, -ImaginaryNumber / magnitude);
        }

        #endregion

        #region 문자열 구하기 - ToString()

        /// <summary>
        /// 문자열 구하기
        /// </summary>
        /// <returns>문자열</returns>
        public override string ToString()
        {
            return $"{RealNumber} + {ImaginaryNumber}i";
        }

        #endregion
    }
}