﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;

namespace TestProject
{
    /// <summary>
    /// 그래픽 확장
    /// </summary>
    public static class GraphicExtension
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 대시 상자 그리기 - DrawDashedBox(graphics, color1, color2, thickness, dashSize, point1, point2)

        /// <summary>
        /// 대시 상자 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="color1">색상 1</param>
        /// <param name="color2">색상 2</param>
        /// <param name="thickness">두께</param>
        /// <param name="dashSize">대시 크기</param>
        /// <param name="point1">포인트 1</param>
        /// <param name="point2">포인트 2</param>
        public static void DrawDashedBox(this Graphics graphics, Color color1, Color color2, float thickness, float dashSize, Point point1, Point point2)
        {
            Rectangle rectangle = point1.ToRectangle(point2);

            using(Pen pen = new Pen(color1, thickness))
            {
                graphics.DrawRectangle(pen, rectangle);

                pen.DashPattern = new float[] { dashSize, dashSize };
                pen.Color       = color2;

                graphics.DrawRectangle(pen, rectangle);
            }
        }

        #endregion
        #region 사각형 구하기 - ToRectangle(point1, point2)

        /// <summary>
        /// 사각형 구하기
        /// </summary>
        /// <param name="point1">포인트 1</param>
        /// <param name="point2">포인트 2</param>
        /// <returns>사각형</returns>
        public static Rectangle ToRectangle(this Point point1, Point point2)
        {
            int x = Math.Min(point1.X, point2.X);
            int y = Math.Min(point1.Y, point2.Y);

            int width  = Math.Abs(point1.X - point2.X);
            int height = Math.Abs(point1.Y - point2.Y);

            return new Rectangle(x, y, width, height);
        }

        #endregion
        #region 사각형 구하기 - ToRectangle(point1, point2)

        /// <summary>
        /// 사각형 구하기
        /// </summary>
        /// <param name="point1">포인트 1</param>
        /// <param name="point2">포인트 2</param>
        /// <returns>사각형</returns>
        public static RectangleF ToRectangle(this PointF point1, PointF point2)
        {
            float x = Math.Min(point1.X, point2.X);
            float y = Math.Min(point1.Y, point2.Y);

            float width  = Math.Abs(point1.X - point2.X);
            float height = Math.Abs(point1.Y - point2.Y);

            return new RectangleF(x, y, width, height);
        }

        #endregion
        #region 이미지 저장하기 - SaveImage(image, filePath)

        /// <summary>
        /// 이미지 저장하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="filePath">파일 경로</param>
        public static void SaveImage(this Image image, string filePath)
        {
            string fileExtension = Path.GetExtension(filePath);

            switch(fileExtension.ToLower())
            {
                case ".bmp"  : image.Save(filePath, ImageFormat.Bmp ); break;
                case ".exif" : image.Save(filePath, ImageFormat.Exif); break;
                case ".gif"  : image.Save(filePath, ImageFormat.Gif ); break;
                case ".jpg"  :
                case ".jpeg" : image.Save(filePath, ImageFormat.Jpeg); break;
                case ".png"  : image.Save(filePath, ImageFormat.Png ); break;
                case ".tif"  :
                case ".tiff" : image.Save(filePath, ImageFormat.Tiff); break;
                default      : throw new NotSupportedException($"알 수 없는 파일 확장자 {fileExtension}");
            }
        }

        #endregion
    }
}