﻿using System;
using System.CodeDom.Compiler;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;

using Microsoft.CSharp;

namespace TestProject
{
    /// <summary>
    /// 메인 폼
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Enumeration
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 스무딩 타입 - SmoothingType

        /// <summary>
        /// 스무딩 타입
        /// </summary>
        private enum SmoothingType
        {
            /// <summary>
            /// None
            /// </summary>
            None,

            /// <summary>
            /// Smooth1
            /// </summary>
            Smooth1,

            /// <summary>
            /// Smooth2
            /// </summary>
            Smooth2,
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 프랙탈 색상 배열
        /// </summary>
        private Color[] fractalColorArray =
        {
            Color.Black,
            Color.Red,
            Color.Orange,
            Color.Yellow,
            Color.Green,
            Color.Cyan,
            Color.Blue,
            Color.Fuchsia
        };

        /// <summary>
        /// 색상 카운트
        /// </summary>
        private int colorCount = 0;

        /// <summary>
        /// 스무딩 타입
        /// </summary>
        private SmoothingType smoothingType = SmoothingType.None;

        /// <summary>
        /// 최대 크기
        /// </summary>
        private int maximumMagnitude;

        /// <summary>
        /// 최대 반복 카운트
        /// </summary>
        private int maximumIterationCount;

        /// <summary>
        /// 로그 단위 탈출 기준
        /// </summary>
        private double logEscape;

        /// <summary>
        /// 테스트 테두리 사각형
        /// </summary>
        private RectangleF testBoundRectangle = RectangleF.FromLTRB(-2.2f, -2.0f, 2.2f, 2.0f);

        /// <summary>
        /// 월드 테두리 사각형
        /// </summary>
        private RectangleF worldBoundRectangle;

        /// <summary>
        /// 시작 포인트
        /// </summary>
        private Point startPoint;

        /// <summary>
        /// 종료 포인트
        /// </summary>
        private Point endPoint;

        /// <summary>
        /// 영역 선택 여부
        /// </summary>
        private bool selectingArea = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainForm()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            Load                                 += Form_Load;
            this.drawButton.Click                += drawButton_Click;
            this.saveFileAsMenuItem.Click        += saveFileAsMenuItem_Click;
            this.exitMenuItem.Click              += exitMenuItem_Click;
            this.noSmoothingMenuItem.Click       += noSmoothingMenuItem_Click;
            this.smooth1MenuItem.Click           += smooth1MenuItem_Click;
            this.smooth2MenuItem.Click           += smooth2MenuItem_Click;
            this.redrawMenuItem.Click            += redrawMenuItem_Click;
            this.scaleOutX2MenuItem.Click        += scaleOutX2MenuItem_Click;
            this.scaleOutX4MenuItem.Click        += scaleOutX4MenuItem_Click;
            this.scaleOutX8MenuItem.Click        += scaleOutX8MenuItem_Click;
            this.fullScaleMenuItem.Click         += fullScaleMenuItem_Click;
            this.enterSelectedAreaMenuItem.Click += enterSelectedAreaMenuItem_Click;
            this.mandelbrotSetMenuItem.Click     += mandelbrotSetMenuItem_Click;
            this.juliaSetBasicMenuItem.Click     += juliaSetBasicMenuItem_Click;
            this.juliaSet1MenuItem.Click         += juliaSet1MenuItem_Click;
            this.juliaSet2MenuItem.Click         += juliaSet2MenuItem_Click;
            this.juliaSet3MenuItem.Click         += juliaSet3MenuItem_Click;
            this.juliaSet4MenuItem.Click         += juliaSet4MenuItem_Click;
            this.juliaSet5MenuItem.Click         += juliaSet5MenuItem_Click;
            this.juliaSet6MenuItem.Click         += juliaSet6MenuItem_Click;
            this.juliaSet7MenuItem.Click         += juliaSet7MenuItem_Click;
            this.juliaSet8MenuItem.Click         += juliaSet8MenuItem_Click;
            this.juliaSet9MenuItem.Click         += juliaSet9MenuItem_Click;
            this.juliaSet10MenuItem.Click        += juliaSet10MenuItem_Click;
            this.juliaSet11MenuItem.Click        += juliaSet11MenuItem_Click;
            this.juliaSet12MenuItem.Click        += juliaSet12MenuItem_Click;
            this.juliaishSet1MenuItem.Click      += juliaishSet1MenuItem_Click;
            this.juliaishSet2MenuItem.Click      += juliaishSet2MenuItem_Click;
            this.juliaishSet3MenuItem.Click      += juliaishSet3MenuItem_Click;
            this.juliaishSet4MenuItem.Click      += juliaishSet4MenuItem_Click;
            this.pictureBox.Paint                += pictureBox_Paint;
            this.pictureBox.MouseDown            += pictureBox_MouseDown;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Event
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 폼 로드시 처리하기 - Form_Load(sender, e)

        /// <summary>
        /// 폼 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Form_Load(object sender, EventArgs e)
        {
            this.pictureBox.Cursor = Cursors.Cross;

            this.colorCount = this.fractalColorArray.Length;

            this.maximumMagnitude = 2;

            this.maximumIterationCount = 64;

            this.logEscape = Math.Log(this.maximumMagnitude);

            SetWorldBoundRectangle(this.testBoundRectangle);

            this.pictureBox.Refresh();
        }

        #endregion

        #region 다른 이름으로 저장하기 메뉴 항목 클릭시 처리하기 - saveFileAsMenuItem_Click(sender, e)

        /// <summary>
        /// 다른 이름으로 저장하기 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void saveFileAsMenuItem_Click(object sender, EventArgs e)
        {
            if(this.pictureBox.Image == null)
            {
                return;
            }

            if(this.saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                this.pictureBox.Image.SaveImage(this.saveFileDialog.FileName);
            }
        }

        #endregion
        #region 종료 메뉴 항목 클릭시 처리하기 - exitMenuItem_Click(sender, e)

        /// <summary>
        /// 종료 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void exitMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        #endregion

        #region No Smoothing 메뉴 항목 클릭시 처리하기 - noSmoothingMenuItem_Click(sender, e)

        /// <summary>
        /// No Smoothing 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void noSmoothingMenuItem_Click(object sender, EventArgs e)
        {
            SelectSmoothing(sender, SmoothingType.None);
        }

        #endregion
        #region Smooth 1 메뉴 항목 클릭시 처리하기 - smooth1MenuItem_Click(sender, e)

        /// <summary>
        /// Smooth 1 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void smooth1MenuItem_Click(object sender, EventArgs e)
        {
            SelectSmoothing(sender, SmoothingType.Smooth1);
        }

        #endregion
        #region Smooth 2 메뉴 항목 클릭시 처리하기 - smooth2MenuItem_Click(sender, e)

        /// <summary>
        /// Smooth 2 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void smooth2MenuItem_Click(object sender, EventArgs e)
        {
            SelectSmoothing(sender, SmoothingType.Smooth2);
        }

        #endregion

        #region 다시 그리기 메뉴 항목 클릭시 처리하기 - redrawMenuItem_Click(sender, e)

        /// <summary>
        /// 다시 그리기 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void redrawMenuItem_Click(object sender, EventArgs e)
        {
            DrawFractal();
        }

        #endregion
        #region 2배 확대 메뉴 항목 클릭시 처리하기 - scaleOutX2MenuItem_Click(sender, e)

        /// <summary>
        /// 2배 확대 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void scaleOutX2MenuItem_Click(object sender, EventArgs e)
        {
            ScaleOut(2);
        }

        #endregion
        #region 4배 확대 메뉴 항목 클릭시 처리하기 - scaleOutX4MenuItem_Click(sender, e)

        /// <summary>
        /// 4배 확대 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void scaleOutX4MenuItem_Click(object sender, EventArgs e)
        {
            ScaleOut(4);
        }

        #endregion
        #region 8배 확대 메뉴 항목 클릭시 처리하기 - scaleOutX8MenuItem_Click(sender, e)

        /// <summary>
        /// 8배 확대 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void scaleOutX8MenuItem_Click(object sender, EventArgs e)
        {
            ScaleOut(8);
        }

        #endregion
        #region 전체 스케일 메뉴 항목 클릭시 처리하기 - fullScaleMenuItem_Click(sender, e)

        /// <summary>
        /// 전체 스케일 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void fullScaleMenuItem_Click(object sender, EventArgs e)
        {
            this.worldBoundRectangle = this.testBoundRectangle;

            DrawFractal();
        }

        #endregion
        #region 선택 영역 입력 메뉴 항목 클릭시 처리하기 - enterSelectedAreaMenuItem_Click(sender, e)

        /// <summary>
        /// 선택 영역 입력 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void enterSelectedAreaMenuItem_Click(object sender, EventArgs e)
        {
            InputForm form = new InputForm();

            form.Text              = "선택 영역";
            form.captionLabel.Text = "최소X, 최소Y, 최대X, 최대Y";
            form.valueTextBox.Text = string.Format
            (
                "{0}, {1}, {2}, {3}",
                this.worldBoundRectangle.Left,
                this.worldBoundRectangle.Top,
                this.worldBoundRectangle.Right,
                this.worldBoundRectangle.Bottom
            );

            if(form.ShowDialog() == DialogResult.OK)
            {
                string[] fieldArray = form.valueTextBox.Text.Split(',');

                float minimumX = float.Parse(fieldArray[0]);
                float minimumY = float.Parse(fieldArray[1]);
                float maximumX = float.Parse(fieldArray[2]);
                float maximumY = float.Parse(fieldArray[3]);

                this.worldBoundRectangle = new RectangleF
                (
                    minimumX,
                    minimumY,
                    maximumX - minimumX,
                    maximumY - minimumY
                );

                DrawFractal();
            }
        }

        #endregion

        #region 픽처 박스 페인트시 처리하기 - pictureBox_Paint(sender, e)

        /// <summary>
        /// 픽처 박스 페인트시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_Paint(object sender, PaintEventArgs e)
        {
            if(this.selectingArea)
            {
                e.Graphics.DrawDashedBox(Color.Red, Color.Yellow, 2, 5, this.startPoint, this.endPoint);
            }
        }

        #endregion
        #region 픽처 박스 마우스 DOWN 처리하기 - pictureBox_MouseDown(sender, e)

        /// <summary>
        /// 픽처 박스 마우스 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_MouseDown(object sender, MouseEventArgs e)
        {
            this.startPoint = e.Location;
            this.endPoint   = e.Location;

            this.pictureBox.MouseMove += pictureBox_MouseMove;
            this.pictureBox.MouseUp   += pictureBox_MouseUp;

            this.selectingArea = true;

            this.pictureBox.Refresh();
        }

        #endregion
        #region 픽처 박스 마우스 이동시 처리하기 - pictureBox_MouseMove(sender, e)

        /// <summary>
        /// 픽처 박스 마우스 이동시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            this.endPoint = e.Location;

            this.pictureBox.Refresh();
        }

        #endregion
        #region 픽처 박스 마우스 UP 처리하기 - pictureBox_MouseUp(sender, e)

        /// <summary>
        /// 픽처 박스 마우스 UP 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_MouseUp(object sender, MouseEventArgs e)
        {
            this.selectingArea = false;

            this.pictureBox.MouseMove -= pictureBox_MouseMove;
            this.pictureBox.MouseUp   -= pictureBox_MouseUp;

            PointF startWorldPoint = GetWorldPoint(this.startPoint);
            PointF endWorldPoint   = GetWorldPoint(this.endPoint  );

            SetWorldBoundRectangle(startWorldPoint.ToRectangle(endWorldPoint));
        }

        #endregion

        #region Mandelbrot Set 메뉴 항목 클릭시 처리하기 - mandelbrotSetMenuItem_Click(sender, e)

        /// <summary>
        /// Mandelbrot Set 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void mandelbrotSetMenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "0";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + x + y * i";

            DrawFractal();
        }

        #endregion
        #region Julia Set Basic 메뉴 항목 클릭시 처리하기 - juliaSetBasicMenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set Basic 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSetBasicMenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) - i";

            DrawFractal();
        }

        #endregion
        #region Julia Set 1 메뉴 항목 클릭시 처리하기 - juliaSet1MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 1 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet1MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (-0.4 + 0.6 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 2 메뉴 항목 클릭시 처리하기 - juliaSet2MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 2 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet2MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (0.285 + 0 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 3 메뉴 항목 클릭시 처리하기 - juliaSet3MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 3 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet3MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (0.285 + 0.01 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 4 메뉴 항목 클릭시 처리하기 - juliaSet4MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 4 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet4MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (0.45 + 0.1428 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 5 메뉴 항목 클릭시 처리하기 - juliaSet5MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 5 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet5MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (-0.70176 - 0.3842 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 6 메뉴 항목 클릭시 처리하기 - juliaSet6MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 6 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet6MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (-0.835 - 0.2321 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 7 메뉴 항목 클릭시 처리하기 - juliaSet7MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 7 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet7MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (-0.8 + 0.156 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 8 메뉴 항목 클릭시 처리하기 - juliaSet8MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 8 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet8MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (-0.7269 + 0.1889 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 9 메뉴 항목 클릭시 처리하기 - juliaSet9MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 9 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet9MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (0.0 - 0.8 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 10 메뉴 항목 클릭시 처리하기 - juliaSet10MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 10 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet10MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (-0.74543 + 0.11301 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 11 메뉴 항목 클릭시 처리하기 - juliaSet11MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 11 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet11MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (-0.75 + 0.11 * i)";

            DrawFractal();
        }

        #endregion
        #region Julia Set 12 메뉴 항목 클릭시 처리하기 - juliaSet12MenuItem_Click(sender, e)

        /// <summary>
        /// Julia Set 12 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaSet12MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "Zn.GetPower(2) + (-0.1 + 0.651 * i)";

            DrawFractal();
        }

        #endregion
        #region Juliaish Set 1 메뉴 항목 클릭시 처리하기 - juliaishSet1MenuItem_Click(sender, e)

        /// <summary>
        /// Juliaish Set 1 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaishSet1MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "(1 - Zn.GetPower(3) / 6) / (Zn - Zn.GetPower(2) / 2).GetPower(2) + 0.15";

            DrawFractal();
        }

        #endregion
        #region Juliaish Set 2 메뉴 항목 클릭시 처리하기 - juliaishSet2MenuItem_Click(sender, e)

        /// <summary>
        /// Juliaish Set 2 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaishSet2MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "1 - Zn.GetPower(2) + Zn.GetPower(5) / (2 + 4 * Zn)";

            DrawFractal();
        }

        #endregion
        #region Juliaish Set 3 메뉴 항목 클릭시 처리하기 - juliaishSet3MenuItem_Click(sender, e)

        /// <summary>
        /// Juliaish Set 3 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaishSet3MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "(1 - Zn.GetPower(2) + Zn.GetPower(5)) / (2 + 4 * Zn)";

            DrawFractal();
        }

        #endregion
        #region Juliaish Set 4 메뉴 항목 클릭시 처리하기 - juliaishSet4MenuItem_Click(sender, e)

        /// <summary>
        /// Juliaish Set 4 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void juliaishSet4MenuItem_Click(object sender, EventArgs e)
        {
            this.z0TextBox.Text      = "x + y * i";
            this.znPlus1TextBox.Text = "1 - Zn.GetPower(2) + Zn.GetPower(5) / (2 + 4 * Zn) + 0.2";

            DrawFractal();
        }

        #endregion
        #region 그리기 버튼 클릭시 처리하기 - drawButton_Click(sender, e)

        /// <summary>
        /// 그리기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void drawButton_Click(object sender, EventArgs e)
        {
            DrawFractal();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Function
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 테두리 사각형 조정하기 - AdjustBoundRectangle()

        /// <summary>
        /// 테두리 사각형 조정하기
        /// </summary>
        private void AdjustBoundRectangle()
        {
            float worldWidth  = this.worldBoundRectangle.Width;
            float worldHeight = this.worldBoundRectangle.Height;

            float worldAspectRatio   = worldWidth / worldHeight;
            float deviceAspectRation = (float)this.pictureBox.ClientSize.Width / (float)this.pictureBox.ClientSize.Height;

            if(worldAspectRatio > deviceAspectRation)
            {
                worldHeight = worldWidth / deviceAspectRation;
            }
            else
            {
                worldWidth = worldHeight * deviceAspectRation;
            }

            float centerX = this.worldBoundRectangle.X + this.worldBoundRectangle.Width  / 2f;
            float centerY = this.worldBoundRectangle.Y + this.worldBoundRectangle.Height / 2f;

            this.worldBoundRectangle = new RectangleF
            (
                centerX - worldWidth  / 2f,
                centerY - worldHeight / 2f,
                worldWidth,
                worldHeight
            );
        }

        #endregion
        #region 월드 테두리 사각형 설정하기 - SetWorldBoundRectangle(boundRectangle)

        /// <summary>
        /// 월드 테두리 사각형 설정하기
        /// </summary>
        /// <param name="boundRectangle">테두리 사각형</param>
        private void SetWorldBoundRectangle(RectangleF boundRectangle)
        {
            this.worldBoundRectangle = boundRectangle;

            AdjustBoundRectangle();

            DrawFractal();
        }

        #endregion

        #region 장치 포인트 구하기 - GetDevicePoint(worldPoint)

        /// <summary>
        /// 장치 포인트 구하기
        /// </summary>
        /// <param name="worldPoint">월드 포인트</param>
        /// <returns>장치 포인트</returns>
        private PointF GetDevicePoint(PointF worldPoint)
        {
            float scaleX = (worldPoint.X - this.worldBoundRectangle.X) / this.worldBoundRectangle.Width;
            float scaleY = (worldPoint.Y - this.worldBoundRectangle.Y) / this.worldBoundRectangle.Height;

            float x = 0 + scaleX * this.pictureBox.ClientSize.Width;
            float y = 0 + scaleY * this.pictureBox.ClientSize.Height;

            return new PointF(x, y);
        }

        #endregion
        #region 월드 포인트 구하기 - GetWorldPoint(devicePoint)

        /// <summary>
        /// 월드 포인트 구하기
        /// </summary>
        /// <param name="devicePoint">장치 포인트</param>
        /// <returns>월드 포인트</returns>
        private PointF GetWorldPoint(PointF devicePoint)
        {
            float scaleX = (devicePoint.X - 0) / this.pictureBox.ClientSize.Width;
            float scaleY = (devicePoint.Y - 0) / this.pictureBox.ClientSize.Height;

            float x = this.worldBoundRectangle.X + scaleX * this.worldBoundRectangle.Width;
            float y = this.worldBoundRectangle.Y + scaleY * this.worldBoundRectangle.Height;

            return new PointF(x, y);
        }

        #endregion

        #region 메뉴 항목 선택하기 - SelectMenuItem(sender, menuItemArray)

        /// <summary>
        /// 메뉴 항목 선택하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="menuItemArray">메뉴 항목 배열</param>
        private void SelectMenuItem(object sender, ToolStripMenuItem[] menuItemArray)
        {
            ToolStripMenuItem selectedMenuItem = sender as ToolStripMenuItem;

            foreach(ToolStripMenuItem menuItem in menuItemArray)
            {
                menuItem.Checked = (menuItem == selectedMenuItem);
            }
        }

        #endregion
        #region 스무딩 선택하기 - SelectSmoothing(sender, smoothingType)

        /// <summary>
        /// 스무딩 선택하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="smoothingType">스무딩 타입</param>
        private void SelectSmoothing(object sender, SmoothingType smoothingType)
        {
            ToolStripMenuItem[] menuItemArray =
            {
                this.noSmoothingMenuItem,
                this.smooth1MenuItem,
                this.smooth2MenuItem,
            };

            SelectMenuItem(sender, menuItemArray);

            this.smoothingType = smoothingType;

            DrawFractal();
        }

        #endregion
        #region 확대하기 - ScaleOut(scale)

        /// <summary>
        /// 확대하기
        /// </summary>
        /// <param name="scale">스케일</param>
        private void ScaleOut(float scale)
        {
            float centerX = this.worldBoundRectangle.X + this.worldBoundRectangle.Width  / 2f;
            float centerY = this.worldBoundRectangle.Y + this.worldBoundRectangle.Height / 2f;

            float width  = this.worldBoundRectangle.Width  * scale;
            float height = this.worldBoundRectangle.Height * scale;

            this.worldBoundRectangle = new RectangleF
            (
                centerX - width / 2,
                centerY - height / 2,
                width,
                height
            );

            DrawFractal();
        }

        #endregion

        #region 메소드 정보 구하기 - GetMethodInfo(maximumStep, maximumMagnitude, z0, zn1)

        /// <summary>
        /// 메소드 정보 구하기
        /// </summary>
        /// <param name="maximumStep">최대 단계</param>
        /// <param name="maximumMagnitude">최대 크기</param>
        /// <param name="z0">Z0</param>
        /// <param name="znPlus1">Zn + 1</param>
        /// <returns>메소드 정보</returns>
        private MethodInfo GetMethodInfo(string maximumStep, string maximumMagnitude, string z0, string znPlus1)
        {
            string functionText = @"
using TestProject;

public static class Evaluator
{
    public static PointInfo EvaluatePoint(float x, float y)
    {
        Complex i = new Complex(0, 1);
        Complex Zn = " + z0 + @";
        int step_num;
        for (step_num = 0; step_num < " + maximumStep + @"; step_num++)
        {
            if (Zn.Magnitude > " + maximumMagnitude + @") break;

            // Calculate Zn+1.
            Complex ZnPlus1 = " + znPlus1 + @";
            Zn = ZnPlus1;
        }
        return new PointInfo(Zn, step_num);
    }
};";

            CompilerParameters parameters = new CompilerParameters();

            parameters.GenerateInMemory   = true;
            parameters.GenerateExecutable = false;

            parameters.ReferencedAssemblies.Add(Assembly.GetExecutingAssembly().Location);

            CSharpCodeProvider codeProvider = new CSharpCodeProvider();

            CompilerResults results = codeProvider.CompileAssemblyFromSource(parameters, functionText);

            if(results.Errors.Count > 0)
            {
                string message = "Error compiling the function.";

                foreach(CompilerError compilerError in results.Errors)
                {
                    message += "\n" + compilerError.ErrorText;
                }

                MessageBox.Show(message, "ERROR", MessageBoxButtons.OK, MessageBoxIcon.Error);

                return null;
            }
            else
            {
                Type evaluatorType = results.CompiledAssembly.GetType("Evaluator");

                MethodInfo methodInfo = evaluatorType.GetMethod("EvaluatePoint");

                return methodInfo;
            }
        }

        #endregion
        #region 색상 구하기 - GetColor(mu)

        /// <summary>
        /// 색상 구하기
        /// </summary>
        /// <param name="mu">MU</param>
        /// <returns>색상</returns>
        private Color GetColor(double mu)
        {
            int color1 = (int)mu;

            double t2 = mu - color1;
            double t1 = 1 - t2;

            color1 = color1 % this.colorCount;

            int color2 = (color1 + 1) % this.colorCount;

            double red   = this.fractalColorArray[color1].R * t1 + this.fractalColorArray[color2].R * t2;
            double green = this.fractalColorArray[color1].G * t1 + this.fractalColorArray[color2].G * t2;
            double blue  = this.fractalColorArray[color1].B * t1 + this.fractalColorArray[color2].B * t2;

            return Color.FromArgb((byte)red, (byte)green, (byte)blue);
        }

        #endregion
        #region 스무스 망델브로 색상 구하기 - GetSmoothMandelbrotColor(z, c, stepNumber)

        /// <summary>
        /// 스무스 망델브로 색상 구하기
        /// </summary>
        /// <param name="z">복소수 Z</param>
        /// <param name="c">복소수 C</param>
        /// <param name="stepNumber">단계 번호</param>
        /// <returns>스무스 망델브로 색상</returns>
        private Color GetSmoothMandelbrotColor(Complex z, Complex c, int stepNumber)
        {
            if(stepNumber == 0)
            {
                return this.fractalColorArray[0];
            }

            for(int i = 0; i < 3; i++)
            {
                z = z * z + c;

                stepNumber++;
            }

            double mu = stepNumber + 1 - Math.Log(Math.Log(z.Magnitude)) / this.logEscape;

            if(this.smoothingType == SmoothingType.Smooth2)
            {
                mu = mu / this.maximumIterationCount * this.colorCount;
            }

            return GetColor(mu);
        }

        #endregion
        #region 픽셀 색상 설정하기 - SetPixelColor(bitmapHelper, indexX, indexY, methodInfo, z, stepNumber)

        /// <summary>
        /// 픽셀 색상 설정하기
        /// </summary>
        /// <param name="bitmapHelper">비트맵 헬퍼</param>
        /// <param name="indexX">인덱스 X</param>
        /// <param name="indexY">인덱스 Y</param>
        /// <param name="methodInfo">메소드 정보</param>
        /// <param name="z">복소수 Z</param>
        /// <param name="stepNumber">단계 번호</param>
        private void SetPixelColor(BitmapHelper bitmapHelper, int indexX, int indexY, MethodInfo methodInfo, Complex z, int stepNumber)
        {
            if(stepNumber >= this.maximumIterationCount)
            {
                stepNumber = 0;
            }

            Color color;

            if(this.smoothingType == SmoothingType.None)
            {
                color = this.fractalColorArray[stepNumber % this.colorCount];
            }
            else
            {
                color = GetSmoothMandelbrotColor(z, new Complex(), stepNumber);
            }

            bitmapHelper.SetPixel(indexX, indexY, color.R, color.G, color.B, 255);
        }

        #endregion
        #region 프랙탈 그리기 - DrawFractal(minimumX, minimumY, deltaX, deltaY, width, height)

        /// <summary>
        /// 프랙탈 그리기
        /// </summary>
        /// <param name="minimumX">최소 X</param>
        /// <param name="minimumY">최소 Y</param>
        /// <param name="deltaX">델타 X</param>
        /// <param name="deltaY">델타 Y</param>
        /// <param name="width">너비</param>
        /// <param name="height">높이</param>
        /// <returns>비트맵</returns>
        private Bitmap DrawFractal(float minimumX, float minimumY, float deltaX, float deltaY, int width, int height)
        {
            MethodInfo methodInfo = GetMethodInfo
            (
                this.maximumStepTextBox.Text,
                this.maximumMagnitudeTextBox.Text,
                this.z0TextBox.Text,
                this.znPlus1TextBox.Text
            );

            if(methodInfo == null)
            {
                return null;
            }

            Bitmap bitmap = new Bitmap(width, height);

            BitmapHelper bitmapHelper = new BitmapHelper(bitmap);

            bitmapHelper.LockBitmap();

            object[] methodParameterArray = new object[2];

            float y = minimumY;

            for(int indexY = 0; indexY < bitmapHelper.Height; indexY++)
            {
                float x = minimumX;

                for(int indexX = 0; indexX < bitmapHelper.Width; indexX++)
                {
                    methodParameterArray[0] = x;
                    methodParameterArray[1] = y;

                    PointInfo pointInfo = (PointInfo)methodInfo.Invoke(null, methodParameterArray);

                    SetPixelColor(bitmapHelper, indexX, indexY, methodInfo, pointInfo.Z, pointInfo.StepNumber);

                    x += deltaX;
                }

                y += deltaY;
            }

            bitmapHelper.UnlockBitmap();

            return bitmap;
        }

        #endregion
        #region 프랙탈 그리기 - DrawFractal()

        /// <summary>
        /// 프랙탈 그리기
        /// </summary>
        private void DrawFractal()
        {
            this.pictureBox.Image = null;

            this.pictureBox.Refresh();

            Cursor = Cursors.WaitCursor;

            AdjustBoundRectangle();

            float deltaX = this.worldBoundRectangle.Width  / this.pictureBox.ClientRectangle.Width;
            float deltaY = this.worldBoundRectangle.Height / this.pictureBox.ClientRectangle.Height;

            float minimumX = this.worldBoundRectangle.X;
            float minimumY = this.worldBoundRectangle.Y;

            string z0       = this.z0TextBox.Text;
            string equation = this.znPlus1TextBox.Text;

            int width  = this.pictureBox.ClientSize.Width;
            int height = this.pictureBox.ClientSize.Height;
            
            Bitmap bitmap = DrawFractal(minimumX, minimumY, deltaX, deltaY, width, height);

            this.pictureBox.Image = bitmap;

            Cursor = Cursors.Default;
        }

        #endregion
    }
}